# -*- coding: utf-8 -*-
'''
Gazelib, a toolkit for gaze analysis.
'''
from gazelib import package
__version__ = package.version  # noqa

from gazelib import containers  # noqa
from gazelib import io  # noqa
from gazelib import legacy  # noqa
# import gazelib.settings  # noqa
from gazelib import statistics  # noqa
# import gazelib.validation  # noqa
# from . import conversion  # Must be imported separately
